import logging

from dbnd import register_config_cls
from dbnd._core.plugin.dbnd_plugins import is_airflow_enabled


logger = logging.getLogger(__name__)


# @dbnd.hookimpl
def dbnd_setup_plugin():
    from dbnd_docker.docker.docker_engine_config import DockerEngineConfig
    from dbnd_docker.docker.docker_task import DockerRunTask

    register_config_cls(DockerEngineConfig)
    register_config_cls(DockerRunTask)
    if is_airflow_enabled():
        from dbnd_docker.kubernetes.kubernetes_engine_config import (
            KubernetesEngineConfig,
        )

        register_config_cls(KubernetesEngineConfig)
    logger.debug("Registered kubernetes plugin")



"""
This file is only used in Airflow 1.10.10 and bellow;
Copied from Airflow 1.10.15 to unify secrets management
"""
import copy
import typing
import uuid

from kubernetes.client import models as k8s


if typing.TYPE_CHECKING:
    from dbnd_airflow.compat.airflow_multi_version_shim import Secret


def to_env_secret(secret: "Secret") -> k8s.V1EnvVar:
    """Stores es environment secret"""
    return k8s.V1EnvVar(
        name=secret.deploy_target,
        value_from=k8s.V1EnvVarSource(
            secret_key_ref=k8s.V1SecretKeySelector(name=secret.secret, key=secret.key)
        ),
    )


def to_env_from_secret(secret: "Secret") -> k8s.V1EnvFromSource:
    """Reads from environment to secret"""
    return k8s.V1EnvFromSource(secret_ref=k8s.V1SecretEnvSource(name=secret.secret))


def to_volume_secret(secret: "Secret") -> typing.Tuple[k8s.V1Volume, k8s.V1VolumeMount]:
    """Converts to volume secret"""
    vol_id = f"secretvol{uuid.uuid4()}"
    volume = k8s.V1Volume(
        name=vol_id, secret=k8s.V1SecretVolumeSource(secret_name=secret.secret)
    )
    # if secret.items:
    #     volume.secret.items = self.items
    return (
        volume,
        k8s.V1VolumeMount(mount_path=secret.deploy_target, name=vol_id, read_only=True),
    )


def attach_to_pod(secret: "Secret", pod: k8s.V1Pod) -> k8s.V1Pod:
    """Attaches to pod"""
    cp_pod = copy.deepcopy(pod)
    if secret.deploy_type == "volume":
        volume, volume_mount = to_volume_secret(secret)
        cp_pod.spec.volumes = pod.spec.volumes or []
        cp_pod.spec.volumes.append(volume)
        cp_pod.spec.containers[0].volume_mounts = (
            pod.spec.containers[0].volume_mounts or []
        )
        cp_pod.spec.containers[0].volume_mounts.append(volume_mount)
    if secret.deploy_type == "env" and secret.key is not None:
        env = to_env_secret(secret)
        cp_pod.spec.containers[0].env = cp_pod.spec.containers[0].env or []
        cp_pod.spec.containers[0].env.append(env)
    if secret.deploy_type == "env" and secret.key is None:
        env_from = to_env_from_secret(secret)
        cp_pod.spec.containers[0].env_from = cp_pod.spec.containers[0].env_from or []
        cp_pod.spec.containers[0].env_from.append(env_from)
    return cp_pod



"""
This is included in Airflow 2.x PodGenerator object, and we can switch to it once we drop Airflow 1.x
"""
import copy

from typing import List, Optional

from kubernetes.client import models as k8s


def reconcile_pods(base_pod: k8s.V1Pod, client_pod: Optional[k8s.V1Pod]) -> k8s.V1Pod:
    """
    :param base_pod: has the base attributes which are overwritten if they exist
        in the client pod and remain if they do not exist in the client_pod
    :type base_pod: k8s.V1Pod
    :param client_pod: the pod that the client wants to create.
    :type client_pod: k8s.V1Pod
    :return: the merged pods

    This can't be done recursively as certain fields some overwritten, and some concatenated.
    """
    if client_pod is None:
        return base_pod

    client_pod_cp = copy.deepcopy(client_pod)
    client_pod_cp.spec = reconcile_specs(base_pod.spec, client_pod_cp.spec)
    client_pod_cp.metadata = reconcile_metadata(
        base_pod.metadata, client_pod_cp.metadata
    )
    client_pod_cp = merge_objects(base_pod, client_pod_cp)

    return client_pod_cp


def reconcile_metadata(base_meta, client_meta):
    """
    Merge kubernetes Metadata objects
    :param base_meta: has the base attributes which are overwritten if they exist
        in the client_meta and remain if they do not exist in the client_meta
    :type base_meta: k8s.V1ObjectMeta
    :param client_meta: the spec that the client wants to create.
    :type client_meta: k8s.V1ObjectMeta
    :return: the merged specs
    """
    if base_meta and not client_meta:
        return base_meta
    if not base_meta and client_meta:
        return client_meta
    elif client_meta and base_meta:
        client_meta.labels = merge_objects(base_meta.labels, client_meta.labels)
        client_meta.annotations = merge_objects(
            base_meta.annotations, client_meta.annotations
        )
        extend_object_field(base_meta, client_meta, "managed_fields")
        extend_object_field(base_meta, client_meta, "finalizers")
        extend_object_field(base_meta, client_meta, "owner_references")
        return merge_objects(base_meta, client_meta)

    return None


def reconcile_specs(
    base_spec: Optional[k8s.V1PodSpec], client_spec: Optional[k8s.V1PodSpec]
) -> Optional[k8s.V1PodSpec]:
    """
    :param base_spec: has the base attributes which are overwritten if they exist
        in the client_spec and remain if they do not exist in the client_spec
    :type base_spec: k8s.V1PodSpec
    :param client_spec: the spec that the client wants to create.
    :type client_spec: k8s.V1PodSpec
    :return: the merged specs
    """
    if base_spec and not client_spec:
        return base_spec
    if not base_spec and client_spec:
        return client_spec
    elif client_spec and base_spec:
        client_spec.containers = reconcile_containers(
            base_spec.containers, client_spec.containers
        )
        merged_spec = extend_object_field(base_spec, client_spec, "init_containers")
        merged_spec = extend_object_field(base_spec, merged_spec, "volumes")
        return merge_objects(base_spec, merged_spec)

    return None


def reconcile_containers(
    base_containers: List[k8s.V1Container], client_containers: List[k8s.V1Container]
) -> List[k8s.V1Container]:
    """
    :param base_containers: has the base attributes which are overwritten if they exist
        in the client_containers and remain if they do not exist in the client_containers
    :type base_containers: List[k8s.V1Container]
    :param client_containers: the containers that the client wants to create.
    :type client_containers: List[k8s.V1Container]
    :return: the merged containers

    The runs recursively over the list of containers.
    """
    if not base_containers:
        return client_containers
    if not client_containers:
        return base_containers

    client_container = client_containers[0]
    base_container = base_containers[0]
    client_container = extend_object_field(
        base_container, client_container, "volume_mounts"
    )
    client_container = extend_object_field(base_container, client_container, "env")
    client_container = extend_object_field(base_container, client_container, "env_from")
    client_container = extend_object_field(base_container, client_container, "ports")
    client_container = extend_object_field(
        base_container, client_container, "volume_devices"
    )
    client_container = merge_objects(base_container, client_container)

    return [client_container] + reconcile_containers(
        base_containers[1:], client_containers[1:]
    )


def merge_objects(base_obj, client_obj):
    """
    :param base_obj: has the base attributes which are overwritten if they exist
        in the client_obj and remain if they do not exist in the client_obj
    :param client_obj: the object that the client wants to create.
    :return: the merged objects
    """
    if not base_obj:
        return client_obj
    if not client_obj:
        return base_obj

    client_obj_cp = copy.deepcopy(client_obj)

    if isinstance(base_obj, dict) and isinstance(client_obj_cp, dict):
        base_obj_cp = copy.deepcopy(base_obj)
        base_obj_cp.update(client_obj_cp)
        return base_obj_cp

    for base_key in base_obj.to_dict().keys():
        base_val = getattr(base_obj, base_key, None)
        if not getattr(client_obj, base_key, None) and base_val:
            if not isinstance(client_obj_cp, dict):
                setattr(client_obj_cp, base_key, base_val)
            else:
                client_obj_cp[base_key] = base_val
    return client_obj_cp


def extend_object_field(base_obj, client_obj, field_name):
    """
    :param base_obj: an object which has a property `field_name` that is a list
    :param client_obj: an object which has a property `field_name` that is a list.
        A copy of this object is returned with `field_name` modified
    :param field_name: the name of the list field
    :type field_name: str
    :return: the client_obj with the property `field_name` being the two properties appended
    """
    client_obj_cp = copy.deepcopy(client_obj)
    base_obj_field = getattr(base_obj, field_name, None)
    client_obj_field = getattr(client_obj, field_name, None)

    if (not isinstance(base_obj_field, list) and base_obj_field is not None) or (
        not isinstance(client_obj_field, list) and client_obj_field is not None
    ):
        raise ValueError("The chosen field must be a list.")

    if not base_obj_field:
        return client_obj_cp
    if not client_obj_field:
        setattr(client_obj_cp, field_name, base_obj_field)
        return client_obj_cp

    appended_fields = base_obj_field + client_obj_field
    setattr(client_obj_cp, field_name, appended_fields)
    return client_obj_cp



from dbnd_airflow.constants import AIRFLOW_ABOVE_10


if AIRFLOW_ABOVE_10:
    from airflow.kubernetes.pod_launcher import PodStatus
else:
    from airflow.contrib.kubernetes.pod_launcher import PodStatus


__all__ = ["PodStatus"]



"""
https://kubernetes.io/docs/concepts/overview/working-with-objects/names/
A client-provided string that refers to an object in a resource URL, such as /api/v1/pods/some-name.

DNS-1123 Rules:
* contain only lowercase alphanumeric characters, '-' or '.'
* start with an alphanumeric character
* end with an alphanumeric character
* if subdomain name - contain no more than 253 characters
* if label name - contain at most 63 characters.
"""
import functools
import typing

from dbnd._core.utils.string_utils import clean_job_name, strip_by


if typing.TYPE_CHECKING:
    from typing import Optional

    from dbnd._core.task_run.task_run import TaskRun

MAX_CLEAN_SUBDOMAIN_NAME_DNS1123_LEN = 253
MAX_CLEAN_LABEL_NAME_DNS1123_LEN = 63


def clean_name_dns1123(value, max_size, postfix=None):
    # type:(str, int, Optional[str]) -> str
    """
    Create a dns-1123 compatible name.

    @param value: the base value to transform
    @param max_size: the maximum length allowed for the output
    @param postfix: optional string to add to end of the result value
    @return: dns-1123 compatible name
    """
    cleaned_value = clean_job_name(
        value=value,
        enabled_characters="-.",
        placeholder="-",
        max_size=max_size,
        postfix=postfix,
    )

    # remove any none alphanumeric characters in the beginning and end of the cleaned value
    return strip_by(lambda c: not c.isalnum(), cleaned_value)


# Create a dns-1123 compatible label name
clean_label_name_dns1123 = functools.partial(
    clean_name_dns1123, max_size=MAX_CLEAN_LABEL_NAME_DNS1123_LEN
)

# Create a dns-1123 compatible subdomain name
clean_subdomain_name_dns1123 = functools.partial(
    clean_name_dns1123, max_size=MAX_CLEAN_SUBDOMAIN_NAME_DNS1123_LEN
)


def create_pod_id(task_run):
    # type:(TaskRun) -> str
    """
    DNS-1123 subdomain name from task_run
    """

    return clean_subdomain_name_dns1123(
        "dbnd.{task_family}.{task_name}".format(
            task_family=task_run.task.task_family, task_name=task_run.task.task_name
        ),
        postfix=".%s" % str(task_run.task_run_uid)[:8],
    )



import logging

from typing import Dict, List

from dbnd import parameter
from dbnd_docker.container_engine_config import ContainerEngineConfig


logger = logging.getLogger(__name__)


class DockerEngineConfig(ContainerEngineConfig):
    _conf__task_family = "docker"

    def get_docker_ctrl(self, task_run):
        from dbnd_docker.docker.docker_task_run_ctrl import LocalDockerRunCtrl

        return LocalDockerRunCtrl(task_run=task_run)

    network = parameter(default=None, description="Docker Network to connect to")[str]

    api_version = parameter(
        description="Remote API version. "
        "Set to ``auto`` to automatically detect the server's version.,"
    )[str]
    docker_url = parameter(
        description="URL of the host running the docker daemon."
    ).value("unix://var/run/docker.sock")

    auto_remove = parameter(
        description="Auto-removal of the container on daemon side when the container's process exits."
    ).value(False)
    force_pull = parameter(
        description="Pull the docker image on every run. Default is False."
    ).value(False)

    cpus = parameter(
        description="""Number of CPUs to assign to the container.
                  This value gets multiplied with 1024. See
                  https://docs.docker.com/engine/reference/run/#cpu-share-constraint"""
    ).value(1.0)

    environment = parameter.c(
        description="Environment variables to set in the container. (templated)"
    )[Dict[str, str]]
    mem_limit = parameter(
        description="""Maximum amount of memory the container can use."
               Either a float value, which represents the limit in bytes,
               or a string like ``128m`` or ``1g``."""
    ).none[str]

    network_mode = parameter(description="Network mode for the container.").none[str]

    tls_ca_cert = parameter(
        description="Path to a PEM-encoded certificate authority "
        "to secure the docker connection."
    ).none[str]

    tls_client_cert = parameter(
        description="Path to the PEM-encoded certificate used "
        "to authenticate docker client."
    ).none[str]

    tls_client_key = parameter(
        description="Path to the PEM-encoded key used " "to authenticate docker client."
    ).none[str]

    tls_hostname = parameter(
        description="Hostname to match against the docker server certificate "
        "or False to disable the check."
    ).none[str]

    tls_ssl_version = parameter(
        description="Version of SSL to use when communicating with docker daemon."
    ).none[str]

    volumes = parameter.c(
        description="List of volumes to mount into the container, "
        "e.g. `['/host/path:/container/path',"
        " '/host/path2:/container/path2:ro']`"
    )[List[str]]

    working_dir = parameter(
        description="Working directory to set on the container "
        "(equivalent to the -w switch the docker client)"
    ).none[str]

    tmp_dir = parameter(
        description="Mount point inside the container "
        "to a temporary directory created on the host "
        "by the operator. The path is also made available via the environment variable"
        " ``AIRFLOW_TMP_DIR`` inside the container."
    ).value("/tmp/airflow")

    docker_conn_id = parameter(description="ID of the Airflow connection to use").none[
        str
    ]
    user = parameter(description="Default user inside the docker container. ").none[str]
    shm_size = parameter(
        description="Size of ``/dev/shm`` in bytes. "
        "The size must be greater than 0. "
        "If omitted uses system default."
    ).none[int]

    # xcom_all = parameter(description="Push all the stdout or just the last line. The default is False (last line).")
    # xcom_push = parameter(
    #     description="Does the stdout will be pushed to the next step using XCom. The default is False.")

    dns = parameter.c(description="Docker custom DNS servers")[List[str]]
    dns_search = parameter.c(description="Docker custom DNS search domain")[List[str]]



import logging
import shlex
import typing

from typing import Optional

from dbnd._core.errors import ParseParameterError
from dbnd_docker.docker_ctrl import DockerRunCtrl
from dbnd_docker.kubernetes.kube_dbnd_client import DbndPodCtrl
from dbnd_docker.kubernetes.kubernetes_engine_config import KubernetesEngineConfig


if typing.TYPE_CHECKING:
    import kubernetes.client.models as k8s

logger = logging.getLogger(__name__)


class KubernetesTaskRunCtrl(DockerRunCtrl):
    def __init__(self, **kwargs):
        super(KubernetesTaskRunCtrl, self).__init__(**kwargs)
        self.pod_ctrl = None  # type: Optional[DbndPodCtrl]

    def docker_run(self):

        # dont' describe in local run, do it in remote run
        self.context.settings.system.describe = False
        cmds = shlex.split(self.task.command)

        kubernetes_config = self.task.docker_engine  # type: KubernetesEngineConfig
        if self.task.image:
            # If the image was described in the task itself, it should override the configuration
            try:
                container_repository, container_tag = self.task.image.split(":")
                kubernetes_config = kubernetes_config.clone(
                    container_repository=container_repository,
                    container_tag=container_tag,
                )
            except ValueError as e:
                # ValueError is received when the task.image string is not formatted correctly
                raise ParseParameterError(
                    "Received image %s is not in image format! Expected repo:tag. Exception: %s"
                    % (self.task.image, str(e))
                )

        pod: "k8s.V1Pod" = kubernetes_config.build_pod(
            cmds=cmds,
            task_run=self.task_run,
            include_system_secrets=self.task.task_is_system,  # include only for driver,and other system tasks
        )
        kube_dbnd = kubernetes_config.build_kube_dbnd()

        self.pod_ctrl = kube_dbnd.get_pod_ctrl(
            pod.metadata.name, pod.metadata.namespace
        )
        try:
            self.pod_ctrl.run_pod(pod=pod, task_run=self.task_run)
            if not kubernetes_config.detach_run:
                self.pod_ctrl.delete_pod()
        except Exception:
            # we should not delete pod on Keyboard interrupt
            # external system will take care of that!
            if not kubernetes_config.detach_run:
                self.pod_ctrl.delete_pod()
            raise

        self.pod_ctrl = None

    def on_kill(self):
        if not self.task.docker_engine.detach_run and self.pod_ctrl:
            self.pod_ctrl.delete_pod()



import datetime
import logging
import shlex
import subprocess
import textwrap
import typing

from os import environ
from typing import Any, Dict, List, Optional

import kubernetes.client.models as k8s
import six
import yaml

from kubernetes.config import ConfigException
from six import PY2

import dbnd_docker

from dbnd import parameter
from dbnd._core.configuration.environ_config import (
    ENV_DBND__ENV_IMAGE,
    ENV_DBND__ENV_MACHINE,
    ENV_DBND__TRACKING,
    ENV_DBND_ENV,
    ENV_DBND_USER,
    get_dbnd_project_config,
)
from dbnd._core.current import is_verbose
from dbnd._core.errors import DatabandConfigError, friendly_error
from dbnd._core.log.logging_utils import set_module_logging_to_debug
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.utils.basics.environ_utils import environ_enabled
from dbnd._core.utils.json_utils import dumps_safe
from dbnd._core.utils.structures import combine_mappings
from dbnd_airflow.constants import AIRFLOW_ABOVE_10, AIRFLOW_VERSION_2
from dbnd_docker.container_engine_config import ContainerEngineConfig
from dbnd_docker.docker.docker_task import DockerRunTask
from dbnd_docker.kubernetes.compat import volume_shims
from dbnd_docker.kubernetes.compat.pod_reconciler import reconcile_pods
from dbnd_docker.kubernetes.compat.secrets_shim import attach_to_pod
from dbnd_docker.kubernetes.compat.volume_shims import attach_volume_mount
from dbnd_docker.kubernetes.dns1123_clean_names import (
    clean_label_name_dns1123,
    create_pod_id,
)
from targets import target
from targets.values import TimeDeltaValueType


if typing.TYPE_CHECKING:
    from dbnd_airflow.compat.airflow_multi_version_shim import Secret

logger = logging.getLogger(__name__)

ENV_DBND_POD_NAME = "DBND__POD_NAME"
ENV_DBND_POD_NAMESPACE = "DBND__POD_NAMESPACE"
ENV_DBND_DOCKER_IMAGE = "DBND__DOCKER_IMAGE"
ENV_DBND_AUTO_REMOVE_POD = "DBND__AUTO_REMOVE_POD"


class PodRetryConfiguration(object):
    RETRY_COUNT = "retry_count"
    RETRY_DELAY = "retry_delay"

    def __init__(self, exit_codes_and_reasons_to_retry_info_map):
        self.exit_codes_and_reasons_to_retry_info_map = (
            exit_codes_and_reasons_to_retry_info_map
        )

    @classmethod
    def from_kube_config(cls, kube_config):
        return cls(kube_config.pod_error_cfg_source_dict)

    def get_retry_count(self, exit_code_or_reason):
        if exit_code_or_reason not in self.exit_codes_and_reasons_to_retry_info_map:
            return None

        return self.exit_codes_and_reasons_to_retry_info_map[exit_code_or_reason][
            self.RETRY_COUNT
        ]

    def get_retry_delay(self, exit_code_or_reason):
        if exit_code_or_reason not in self.exit_codes_and_reasons_to_retry_info_map:
            return 0

        return self.exit_codes_and_reasons_to_retry_info_map[exit_code_or_reason][
            self.RETRY_DELAY
        ]

    @property
    def reasons_and_exit_codes(self):
        return self.exit_codes_and_reasons_to_retry_info_map.keys()


class KubernetesEngineConfig(ContainerEngineConfig):
    _conf__task_family = "kubernetes"

    cluster_context = parameter.none().help("Kubernetes cluster context")[str]
    config_file = parameter.none().help("Custom Kubernetes config file")[str]

    in_cluster = parameter(default=None).help(
        "If None, we set it dynamically, according to where we run."
    )[bool]

    image_pull_policy = parameter.value(
        "IfNotPresent", description="Kubernetes image_pull_policy flag"
    )

    image_pull_secrets = parameter.none().help("Secret to use for image pull")[str]
    keep_finished_pods = parameter(default=False).help(
        "Don't delete pods on completion"
    )[bool]
    keep_failed_pods = parameter(default=False).help("Don't delete failed pods")[bool]

    namespace = parameter(default="default")[str]
    secrets = parameter(empty_default=True).help(
        "User secrets to be added to every created pod"
    )[List]
    system_secrets = parameter(empty_default=True).help(
        "System secrets (used by Databand Framework)"
    )[List]
    env_vars = parameter(empty_default=True)[Dict]

    node_selectors = parameter(empty_default=True)[Dict]
    annotations = parameter(empty_default=True)[Dict]
    pods_creation_batch_size = parameter.value(10)[int]
    service_account_name = parameter.none()[str]
    gcp_service_account_keys = parameter.none()[
        str
    ]  # it's actually dict, but KubeConf expects str
    affinity = parameter(empty_default=True)[Dict]
    tolerations = parameter(empty_default=True)[List]

    hostnetwork = parameter.value(False)
    configmaps = parameter(empty_default=True)[List[str]]

    volumes = parameter.none()[List[str]]
    volume_mounts = parameter.none()[List[str]]
    security_context = parameter.none()[List[str]]
    labels = parameter.none()[Dict]

    request_memory = parameter.none()[str]
    request_cpu = parameter.none()[str]
    limit_memory = parameter.none()[str]
    limit_cpu = parameter.none()[str]

    requests = parameter.none()[Dict]
    limits = parameter.none()[Dict]

    pod_error_cfg_source_dict = parameter(
        description="Values for pod error handling configuration"
    )[Dict]

    pod_default_retry_delay = parameter(
        description="The default amount of time to wait between retries of pods",
        default="10s",
    )[datetime.timedelta]

    submit_termination_grace_period = parameter(
        description="timedelta to let the submitted pod enter a final state"
    )[datetime.timedelta]

    startup_timeout_seconds = parameter.value(120)
    show_pod_log = parameter(default=False).help(
        "When using this engine as the task_engine, run tasks sequentially and stream their logs"
    )[bool]
    debug = parameter(default=False).help(
        "Equalent to show_pod_log=True + show all debug information"
    )[bool]
    debug_with_command = parameter(default="").help(
        "Use this command as a pod command instead of the original, can help debug complicated issues"
    )[str]
    debug_phase = parameter(default="").help(
        "Debug mode for speicific phase of pod events. All these events will be printed with the full response from k8s"
    )[str]

    prefix_remote_log = parameter(default=True).help(
        "Adds [driver] or [<task_name>] prefix to logs streamed from Kubernetes to the local log"
    )
    check_unschedulable_condition = parameter(default=True).help(
        "Try to detect non-transient issues that prevent the pod from being scheduled and fail the run if needed"
    )
    check_image_pull_errors = parameter(default=True).help(
        "Try to detect image pull issues that prevent the pod from being scheduled and fail the run if needed"
    )
    check_running_pod_errors = parameter(default=False).help(
        "Try to detect running pod issues like failed ContainersReady condition (pod is deleted)"
    )
    check_cluster_resource_capacity = parameter(default=True).help(
        "When a pod can't be scheduled due to cpu or memory constraints, check if the constraints are possible to satisfy in the cluster"
    )

    startup_timeout = parameter(default="10m").help(
        "Time to wait for pod getting into Running state"
    )[datetime.timedelta]

    dashboard_url = parameter(default=None).help(
        "skeleton url to display as kubernetes dashboard"
    )[str]

    pod_log_url = parameter(default=None).help("skeleton url to display logs of pods")[
        str
    ]

    pod_yaml = parameter(default="${DBND_LIB}/conf/kubernetes-pod.yaml").help(
        "Base yaml to use to run databand task/driver"
    )[str]

    trap_exit_file_flag = parameter(default=None).help("trap exit file")[str]
    auto_remove = parameter(
        default=False,
        description="Auto-removal of the pod when container has finished.",
    )[bool]
    detach_run = parameter(
        default=False, description="Submit run only, do not wait for it completion."
    )[bool]

    watcher_request_timeout_seconds = parameter(
        default=300,
        description="How many seconds watcher should wait "
        "for events "
        "until timeout",
    )[int]

    watcher_recreation_interval_seconds = parameter(
        default=30,
        description="How many seconds to wait before resurrecting watcher after the timeout",
    )[int]

    watcher_client_timeout_seconds = parameter(
        default=50,
        description="How many seconds to wait before timeout occurs in watcher on client side (read)",
    )[int]

    log_pod_events_on_sigterm = parameter(
        default=False,
        description="When receiving sigterm log current pod state to debug why the pod was terminated",
    )

    pending_zombies_timeout = parameter(
        default="5h",
        description="Amount of time we will wait before a pending pod would consider a"
        " zombie and we will set it to fail",
    ).type(TimeDeltaValueType)

    zombie_query_interval_secs = parameter(
        default=600,
        description="Amount of seconds we wait between zombie checking intervals. "
        "Default: 600 sec => 10 minutes",
    )

    zombie_threshold_secs = parameter(
        default=300,
        description="If the job has not heartbeat in this many seconds, "
        "the scheduler will mark the associated task instance as failed and will re-schedule the task.",
    )

    # airflow live logs feature
    # ------------------------- #
    airflow_log_enabled = parameter(
        default=False,
        description="Enables Airflow Live log at KubernetesExecutor feature",
    )[bool]
    airflow_log_image = parameter(
        default=None,
        description="Overrider the image that will be used to add sidecar to the run which will expose the live "
        "logs of the run. By default the main container image will be used",
    )[str]
    airflow_log_folder = parameter(
        default="/usr/local/airflow/logs",
        description="Specify the location on the airflow image (sidecar), where we mount the logs from the original"
        " container and expose them to airflow ui.",
    )[str]
    airflow_log_port = parameter(
        default="8793",
        description="The port airflow live log sidecar will expose its service. This port should match the port "
        "airflow webserver tries to access the live logs ",
    )[str]

    airflow_log_trap_exit_flag_default = parameter(
        default="/tmp/pod/terminated",
        description="The path that will be used by default if `airflow_log_enabled` is true ",
    )[str]

    container_airflow_log_path = parameter(
        default="/root/airflow/logs/",
        description="The path to the airflow logs, on the databand container.",
    )
    host_as_ip_for_live_logs = parameter(
        default=True,
        description="Set the host of the pod to be the ip address of the pod."
        "In kubernetes normally, only Services get DNS names, not Pods."
        "We use the ip for airflow webserver to lookup the sidecar See more here: https://stackoverflow.com/a/59262628",
    )

    def _initialize(self):
        super(KubernetesEngineConfig, self)._initialize()

        if self.debug:
            logger.warning(
                "Running in debug mode, setting all k8s loggers to debug, waiting for every pod completion!"
            )
            if AIRFLOW_VERSION_2:
                from airflow import kubernetes
            else:
                from airflow.contrib import kubernetes

            set_module_logging_to_debug([dbnd_docker, kubernetes])
            self.detach_run = False
        if self.show_pod_log:
            logger.warning(
                "Showing pod logs at runtime, waiting for every pod completion!"
            )
            self.detach_run = False
        if self.auto_remove and not self.detach_run:
            logger.warning(
                "Can't auto remove pod if not running from detach_run=True mode, "
                "switching to auto_remove=False"
            )
            self.auto_remove = False

        if (
            self.airflow_log_enabled
            and not self.trap_exit_file_flag
            and self.airflow_log_trap_exit_flag_default
        ):
            self.trap_exit_file_flag = self.airflow_log_trap_exit_flag_default

        self.pod_retry_config = PodRetryConfiguration.from_kube_config(self)

    def get_docker_ctrl(self, task_run):
        from dbnd_docker.kubernetes.kubernetes_task_run_ctrl import (
            KubernetesTaskRunCtrl,
        )

        return KubernetesTaskRunCtrl(task_run=task_run)

    def submit_to_engine_task(self, env, task_name, args, interactive=True):
        docker_engine = self
        if not interactive:
            docker_engine = docker_engine.clone(auto_remove=True, detach_run=True)
        return DockerRunTask(
            task_name=task_name,
            command=subprocess.list2cmdline(args),
            image=self.full_image,
            docker_engine=docker_engine,
            task_is_system=True,
        )

    def cleanup_after_run(self):
        # this run was submitted by task_run_async - we need to cleanup ourself
        if not environ_enabled(ENV_DBND_AUTO_REMOVE_POD):
            return
        if ENV_DBND_POD_NAME in environ and ENV_DBND_POD_NAMESPACE in environ:
            try:
                logger.warning(
                    "Auto deleteing pod as accordingly to '%s' env variable"
                    % ENV_DBND_AUTO_REMOVE_POD
                )
                kube_dbnd = self.build_kube_dbnd()
                kube_dbnd.delete_pod(
                    name=environ[ENV_DBND_POD_NAME],
                    namespace=environ[ENV_DBND_POD_NAMESPACE],
                )
            except Exception as e:
                logger.warning("Tried to delete this pod but failed: %s" % e)
        else:
            logger.warning(
                "Auto deleting pod as set, but pod name and pod namespace is not defined"
            )

    def get_dashboard_link(self, pod_namespace: str, pod_name: str) -> Optional[str]:
        if not self.dashboard_url:
            return None
        try:
            return self.dashboard_url.format(namespace=pod_namespace, pod=pod_name)
        except Exception:
            logger.exception(
                "Failed to generate dashboard url from %s" % self.dashboard_url
            )
        return None

    def get_pod_log_link(self, pod_namespace: str, pod_name: str) -> Optional[str]:
        if not self.pod_log_url:
            return None
        try:
            return self.pod_log_url.format(
                namespace=pod_namespace,
                pod=pod_name,
                timestamp=datetime.datetime.now().isoformat(),
            )
        except Exception:
            logger.exception("Internal error on generating pod log url")
        return None

    def get_kube_client(self, in_cluster=None):
        from kubernetes import client, config

        # if in_cluster is set to None, we set it dynamically by trying to set
        # the k8s's config as if we are in a cluster, and if it fails, we set it
        # as we are not running in a cluster.
        if in_cluster is None:
            try:
                config.load_incluster_config()
                self.in_cluster = True
            except ConfigException:
                try:
                    config.load_kube_config(
                        config_file=self.config_file, context=self.cluster_context
                    )
                except ConfigException as e:
                    raise friendly_error.executor_k8s.failed_to_load_config_file(e)
                self.in_cluster = False
        else:
            try:
                if in_cluster:
                    config.load_incluster_config()
                else:
                    config.load_kube_config(
                        config_file=self.config_file, context=self.cluster_context
                    )
            except ConfigException as e:
                raise friendly_error.executor_k8s.failed_to_connect_to_cluster(
                    self.in_cluster, e
                )

        if PY2:
            # For connect_get_namespaced_pod_exec
            from kubernetes.client import Configuration

            configuration = Configuration()
            configuration.assert_hostname = False
            Configuration.set_default(configuration)
        return client.CoreV1Api()

    def build_kube_dbnd(self, in_cluster=None):
        from dbnd_docker.kubernetes.kube_dbnd_client import DbndKubernetesClient

        kube_client = self.get_kube_client(in_cluster=in_cluster)
        kube_dbnd = DbndKubernetesClient(kube_client=kube_client, engine_config=self)
        return kube_dbnd

    def get_pod_name(self, task_run, try_number):
        pod_name = create_pod_id(task_run)
        if try_number is not None:
            pod_name = "%s-%s" % (pod_name, try_number)
        return pod_name

    def build_pod(
        self,
        task_run: TaskRun,
        cmds: List[str],
        args: Optional[List[str]] = None,
        labels: Optional[Dict[str, str]] = None,
        try_number: Optional[int] = None,
        include_system_secrets: bool = False,
    ) -> k8s.V1Pod:
        if not self.container_tag:
            raise DatabandConfigError(
                "Your container tag is None, please check your configuration",
                help_msg="Container tag should be assigned",
            )

        pod_name = self.get_pod_name(task_run=task_run, try_number=try_number)

        image = self.full_image
        labels = combine_mappings(labels, self.labels)
        labels["pod_name"] = pod_name

        labels["dbnd_run_uid"] = task_run.run.run_uid
        labels["dbnd_task_run_uid"] = task_run.task_run_uid
        labels["dbnd_task_run_attempt_uid"] = task_run.task_run_attempt_uid
        labels[
            "dbnd_task_family"
        ] = task_run.task.task_definition.full_task_family_short
        labels["dbnd_task_name"] = task_run.task.task_name
        labels["dbnd_task_af_id"] = task_run.task_af_id

        # for easier pod deletion (kubectl delete pod -l dbnd=task_run -n <my_namespace>)
        if task_run.task.task_is_system:
            labels["dbnd"] = "dbnd_system_task_run"
        else:
            labels["dbnd"] = "task_run"

        # we need to be sure that the values meet the dns label names RFC
        # https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#dns-label-names
        labels = {
            label_name: clean_label_name_dns1123(str(label_value))
            for label_name, label_value in six.iteritems(labels)
        }
        if is_verbose():
            logger.info("Build pod with kubernetes labels {}".format(labels))

        annotations = self.annotations.copy()
        if self.gcp_service_account_keys:
            annotations[
                "iam.cloud.google.com/service-account"
            ] = self.gcp_service_account_keys
        annotations["dbnd_tracker"] = task_run.task_tracker_url

        from dbnd_docker.kubernetes.vendorized_airflow.dbnd_extended_resources import (
            DbndExtendedResources,
        )

        resources = DbndExtendedResources(
            requests=self.requests,
            limits=self.limits,
            request_memory=self.request_memory,
            request_cpu=self.request_cpu,
            limit_memory=self.limit_memory,
            limit_cpu=self.limit_cpu,
        )
        env_vars = {
            ENV_DBND_POD_NAME: pod_name,
            ENV_DBND_POD_NAMESPACE: self.namespace,
            ENV_DBND_USER: task_run.task_run_env.user,
            ENV_DBND__ENV_IMAGE: image,
            ENV_DBND_ENV: task_run.run.env.task_name,
            ENV_DBND__ENV_MACHINE: "%s at %s" % (pod_name, self.namespace),
        }

        if AIRFLOW_VERSION_2:
            env_vars[
                "AIRFLOW__CORE__TASK_RUNNER"
            ] = "dbnd_airflow.compat.dbnd_task_runner.DbndStandardTaskRunner"

        if self.auto_remove:
            env_vars[ENV_DBND_AUTO_REMOVE_POD] = "True"
        env_vars[self._params.get_param_env_key(self, "in_cluster")] = "True"
        env_vars["AIRFLOW__KUBERNETES__IN_CLUSTER"] = "True"
        env_vars[
            "DBND__RUN_INFO__SOURCE_VERSION"
        ] = task_run.run.context.task_run_env.user_code_version
        env_vars["AIRFLOW__KUBERNETES__DAGS_IN_IMAGE"] = "True"
        if not get_dbnd_project_config().is_tracking_mode():
            env_vars[ENV_DBND__TRACKING] = "False"
        # we want that all next runs will be able to use the image that we have in our configuration

        env_vars.update(
            self._params.to_env_map(self, "container_repository", "container_tag")
        )

        env_vars.update(self.env_vars)
        env_vars.update(task_run.run.get_context_spawn_env())

        secrets = self.get_secrets(include_system_secrets=include_system_secrets)

        if self.trap_exit_file_flag:
            args = [
                textwrap.dedent(
                    """
                trap "touch {trap_file}" EXIT
                {command}
                """.format(
                        trap_file=self.trap_exit_file_flag,
                        command=subprocess.list2cmdline(cmds),
                    )
                )
            ]
            # we update cmd now
            cmds = ["/bin/bash", "-c"]

        if self.debug_with_command:
            logger.warning(
                "%s replacing pod %s command with '%s', original command=`%s`",
                task_run,
                pod_name,
                self.debug_with_command,
                subprocess.list2cmdline(cmds),
            )
            cmds = shlex.split(self.debug_with_command)

        base_pod = self._build_base_pod()

        pod = self._to_real_pod(
            cmds=cmds,
            args=args,
            namespace=self.namespace,
            name=pod_name,
            envs=env_vars,
            image=image,
            labels=labels,
            secrets=secrets,
            resources=resources,
            annotations=annotations,
        )

        final_pod = reconcile_pods(base_pod, pod)

        return final_pod

    def _to_real_pod(
        self,
        cmds: List[str],
        args: List[str],
        namespace: str,
        name: str,
        image: str,
        envs: Dict[str, str],
        labels: Dict[str, str],
        annotations: Dict[str, str],
        resources: "DbndExtendedResources",
        secrets: List["Secret"],
    ) -> k8s.V1Pod:

        # TODO add yaml template as basis
        BASE_CONTAINER_NAME = "base"
        kc: KubernetesEngineConfig = self
        meta = k8s.V1ObjectMeta(
            labels=labels, name=name, namespace=namespace, annotations=annotations
        )
        if kc.image_pull_secrets:
            image_pull_secrets = [
                k8s.V1LocalObjectReference(i) for i in kc.image_pull_secrets.split(",")
            ]
        else:
            image_pull_secrets = []
        spec = k8s.V1PodSpec(
            # init_containers=kc.init_containers,
            containers=[
                k8s.V1Container(
                    image=image,
                    command=cmds,
                    env_from=[],
                    name=BASE_CONTAINER_NAME,
                    env=[
                        k8s.V1EnvVar(name=key, value=val) for key, val in envs.items()
                    ],
                    args=args,
                    image_pull_policy=kc.image_pull_policy,
                )
            ],
            image_pull_secrets=image_pull_secrets,
            service_account_name=kc.service_account_name,
            node_selector=kc.node_selectors,
            # dns_policy=kc.dnspolicy,
            host_network=kc.hostnetwork,
            tolerations=kc.tolerations,
            affinity=kc.affinity,
            security_context=kc.security_context,
        )

        k8_pod = k8s.V1Pod(spec=spec, metadata=meta)
        for configmap_name in kc.configmaps:
            env_var = k8s.V1EnvFromSource(
                config_map_ref=k8s.V1ConfigMapEnvSource(name=configmap_name)
            )
            k8_pod.spec.containers[0].env_from.append(env_var)

        volumes = kc.volumes or []
        for volume in volumes:
            k8_pod = volume_shims.attach_to_pod(k8_pod, volume)

        mounts = kc.volume_mounts or []
        for volume_mount in mounts:
            k8_pod = attach_volume_mount(k8_pod, volume_mount)

        secret: Secret
        for secret in secrets:
            if AIRFLOW_ABOVE_10:
                k8_pod = secret.attach_to_pod(k8_pod)
            else:
                k8_pod = attach_to_pod(secret, k8_pod)

        k8_pod = resources.attach_to_pod(k8_pod)
        return k8_pod

    def _build_base_pod(self) -> k8s.V1Pod:
        from kubernetes.client import ApiClient

        basis_pod_yaml = target(self.pod_yaml).read()
        basis_pod_dict = yaml.safe_load(basis_pod_yaml) or {}
        api_client = ApiClient()
        return api_client._ApiClient__deserialize_model(basis_pod_dict, k8s.V1Pod)

    def get_secrets(self, include_system_secrets=True) -> List["Secret"]:
        """Defines any necessary secrets for the pod executor"""

        from dbnd_airflow.compat.airflow_multi_version_shim import Secret

        result = []
        if include_system_secrets:
            secrets = self.system_secrets + self.secrets
        else:
            secrets = self.secrets
        for secret_data in secrets:
            result.append(
                Secret(
                    deploy_type=secret_data.get("type"),
                    deploy_target=secret_data.get("target"),
                    secret=secret_data.get("secret"),
                    key=secret_data.get("key"),
                )
            )

        return result

    def build_kube_pod_req(self, pod: k8s.V1Pod) -> Dict[str, Any]:
        from kubernetes.client import ApiClient

        return ApiClient().sanitize_for_serialization(pod)

    # TODO: [#2] add them in-place?
    def apply_env_vars_to_pod(self, pod):
        pod.envs["AIRFLOW__KUBERNETES__DAGS_IN_IMAGE"] = "True"
        if not get_dbnd_project_config().is_tracking_mode():
            pod.envs[ENV_DBND__TRACKING] = "False"


def readable_pod_request(pod_req):
    try:
        return yaml.dump(
            pod_req,
            encoding="utf-8",
            allow_unicode=True,
            sort_keys=False,
            default_flow_style=False,
        ).decode("ascii", errors="ignore")
    except Exception as ex:
        logger.info("Failed to create readable pod request representation: %s", ex)
        return dumps_safe(pod_req)



from __future__ import absolute_import

import logging

from dbnd import parameter
from dbnd.tasks import Task
from dbnd_docker.container_engine_config import ContainerEngineConfig
from dbnd_docker.docker_ctrl import DockerRunCtrl


logger = logging.getLogger(__name__)


class DockerRunTask(Task):
    # _conf__task_family = "docker_cmd"

    image = parameter(
        description="Docker image from which to create the container."
        "If image tag is omitted, 'latest' will be used."
    )[str]
    command = parameter(description="Command to be run in the container. (templated)")[
        str
    ]

    docker_engine = parameter(from_task_env_config=True)[ContainerEngineConfig]

    docker_ctrl = None  # type: DockerRunCtrl

    def _task_submit(self):
        if hasattr(self.ctrl, "airflow_op"):
            airflow_context = self.current_task_run.airflow_context
            self.command = self.ctrl.airflow_op.render_template(
                self.command, airflow_context
            )

        self.log_metric("docker command", self.command)
        self.docker_ctrl = self.docker_engine.get_docker_ctrl(
            self.current_task_run
        )  # type: DockerRunCtrl
        self.docker_ctrl.docker_run()

    def on_kill(self):
        if self.docker_ctrl is not None:
            logger.error("Killing submitted docker for %s", self.task_id)
            return self.docker_ctrl.on_kill()



from dbnd_docker.docker_ctrl import DockerRunCtrl


class LocalDockerRunCtrl(DockerRunCtrl):
    def __init__(self, **kwargs):
        super(LocalDockerRunCtrl, self).__init__(**kwargs)
        self.runner_op = None

    @property
    def docker_config(self):
        # type: (LocalDockerRunCtrl) -> DockerEngineConfig
        return self.task.docker_engine

    def docker_run(self):
        from dbnd_airflow_contrib.docker_operator import DockerOperator

        t = self.task
        dc = self.docker_config
        environment = dc.environment.copy()
        environment.update(self.task_run.run.get_context_spawn_env())

        self.runner_op = DockerOperator(
            task_id=self.task.task_id,
            image=t.image,
            command=t.command,
            docker_conn_id=dc.docker_conn_id,
            cpus=dc.cpus,
            environment=environment,
            api_version=dc.api_version,
            docker_url="unix://var/run/docker.sock",
            force_pull=dc.force_pull,
            mem_limit=dc.mem_limit,
            network_mode=dc.network_mode,
            tls_ca_cert=dc.tls_ca_cert,
            tls_client_cert=dc.tls_client_cert,
            tls_client_key=dc.tls_client_key,
            tls_hostname=dc.tls_hostname,
            tls_ssl_version=dc.tls_ssl_version,
            tmp_dir=dc.tmp_dir,
            user=dc.user,
            volumes=dc.volumes,
            working_dir=dc.working_dir,
            xcom_push=False,
            xcom_all=False,
            dns=dc.dns,
            dns_search=dc.dns_search,
            auto_remove=dc.auto_remove,
            shm_size=dc.shm_size,
        )

        self.runner_op.execute(context=None)

    def on_kill(self):
        if self.runner_op is not None:
            self.runner_op.on_kill()



import copy
import typing

from kubernetes.client import models as k8s


def to_k8s_client_obj(volume_spec: typing.Dict[str, typing.Any]) -> k8s.V1Volume:
    name = volume_spec.get("name")

    volume = k8s.V1Volume(name=name)
    for k, v in volume_spec.items():
        snake_key = _convert_to_snake_case(k)
        if hasattr(volume, snake_key):
            setattr(volume, snake_key, v)
        else:
            raise AttributeError("k8s.V1Volume does not have attribute {}".format(k))
    return volume


def attach_to_pod(
    pod: k8s.V1Pod, volume_spec: typing.Dict[str, typing.Any]
) -> k8s.V1Pod:
    cp_pod = copy.deepcopy(pod)
    volume = to_k8s_client_obj(volume_spec)
    cp_pod.spec.volumes = pod.spec.volumes or []
    cp_pod.spec.volumes.append(volume)
    return cp_pod


def _convert_to_snake_case(some_string):
    return "".join(["_" + i.lower() if i.isupper() else i for i in some_string]).lstrip(
        "_"
    )


def attach_volume_mount(
    pod: k8s.V1Pod, volume_mount_spec: typing.Dict[str, typing.Any]
):
    cp_pod = copy.deepcopy(pod)

    volume_mount = k8s.V1VolumeMount(
        name=volume_mount_spec.get("name"),
        mount_path=volume_mount_spec.get("mountPath"),
        sub_path=volume_mount_spec.get("subPath"),
        read_only=volume_mount_spec.get("readOnly"),
    )
    cp_pod.spec.containers[0].volume_mounts = pod.spec.containers[0].volume_mounts or []
    cp_pod.spec.containers[0].volume_mounts.append(volume_mount)
    return cp_pod



import logging
import pprint
import time
import typing

from datetime import datetime
from typing import Optional

from kubernetes import client
from kubernetes.client.rest import ApiException

from dbnd._core.constants import TaskRunState
from dbnd._core.current import try_get_databand_run
from dbnd._core.errors import DatabandError, DatabandRuntimeError, friendly_error
from dbnd._core.log.logging_utils import PrefixLoggerAdapter, override_log_formatting
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.utils.timezone import utcnow
from dbnd_docker.kubernetes.compat.pod_launcher import PodStatus
from dbnd_docker.kubernetes.kube_resources_checker import DbndKubeResourcesChecker
from dbnd_docker.kubernetes.kubernetes_engine_config import (
    KubernetesEngineConfig,
    readable_pod_request,
)


if typing.TYPE_CHECKING:
    import kubernetes.client.models as k8s

    from kubernetes.client import CoreV1Api, V1Pod

    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


class PodFailureReason(object):
    err_image_pull = "err_image_pull"
    err_config_error = "err_config_error"
    err_pod_deleted = "err_pod_deleted"
    err_pod_evicted = "err_pod_evicted"


class DbndKubernetesClient(object):
    def __init__(self, kube_client, engine_config):
        # type:(DbndKubernetesClient, CoreV1Api,KubernetesEngineConfig)->None
        super(DbndKubernetesClient, self).__init__()

        self.kube_client = kube_client
        self.engine_config = engine_config

    def get_pod_ctrl(self, name, namespace=None, config=None):
        # type: (str, Optional[str], Optional[KubernetesEngineConfig])-> DbndPodCtrl
        return DbndPodCtrl(
            pod_name=name,
            pod_namespace=namespace or self.engine_config.namespace,
            kube_client=self.kube_client,
            kube_config=config or self.engine_config,
        )

    def delete_pod(self, name, namespace):
        self.get_pod_ctrl(name=name, namespace=namespace).delete_pod()

    def get_pod_status(self, pod_name):
        # type: (str) -> Optional[V1Pod]
        return self.get_pod_ctrl(name=pod_name).get_pod_status_v1()


class DbndPodCtrl(object):
    def __init__(self, pod_name, pod_namespace, kube_config, kube_client):
        self.kube_config = kube_config  # type: KubernetesEngineConfig
        self.name = pod_name
        self.namespace = pod_namespace
        self.kube_client = kube_client
        self.log = PrefixLoggerAdapter("pod %s" % self.name, logger)

    def delete_pod(self):
        if self.kube_config.keep_finished_pods:
            self.log.warning("Will not delete pod due to keep_finished_pods=True.")
            return

        if self.kube_config.keep_failed_pods:
            pod_phase = self.get_pod_phase()
            if pod_phase not in {PodPhase.RUNNING, PodPhase.SUCCEEDED}:
                self.log.warning(
                    "Keeping failed pod due to keep_failed_pods=True and state is %s",
                    pod_phase,
                )
                return

        try:
            self.kube_client.delete_namespaced_pod(
                self.name, self.namespace, body=client.V1DeleteOptions()
            )
            self.log.info("Pod has been deleted.")
        except ApiException as e:
            self.log.info(
                "Failed to delete pod: %s", e if e.status != 404 else "pod not found"
            )
            # If the pod is already deleted, don't raise
            # if e.status != 404:
            #     raise

    def get_pod_status_v1(self):
        # type: () -> Optional[V1Pod]
        try:
            return self.kube_client.read_namespaced_pod(
                name=self.name, namespace=self.namespace
            )
        except ApiException as e:
            # If the pod can not be found
            if e.status == 404:
                return None
            raise

    def get_pod_phase(self):
        pod_resp = self.get_pod_status_v1()
        if not pod_resp:
            return None

        return pod_resp.status.phase

    def _wait_for_pod_started(self, _logger=None):
        """
        will try to raise an exception if the pod fails to start (see DbndPodLauncher.check_deploy_errors)
        """
        _logger = _logger or self.log
        start_time = datetime.now()
        while True:
            pod_status = self.get_pod_status_v1()
            if not pod_status:
                raise DatabandError("Can not find pod at k8s:%s")
            # PATCH:  validate deploy errors
            self.check_deploy_errors(pod_status)

            pod_phase = pod_status.status.phase
            if pod_phase.lower() != PodStatus.PENDING:
                return

            startup_delta = datetime.now() - start_time
            if startup_delta >= self.kube_config.startup_timeout:
                raise DatabandError("Pod is still not running after %s" % startup_delta)
            time.sleep(1)
            _logger.debug("Pod not yet started: %s", pod_status.status)

    def stream_pod_logs(self, print_func=logger.info, follow=False, tail_lines=None):
        kwargs = {
            "name": self.name,
            "namespace": self.namespace,
            "container": "base",
            "follow": follow,
            "_preload_content": False,
        }
        if tail_lines:
            kwargs["tail_lines"] = tail_lines

        logs = self.kube_client.read_namespaced_pod_log(**kwargs)
        try:
            if self.kube_config.prefix_remote_log:
                # we want to remove regular header in log, and make it looks like '[pod_name] LOG FROM POD'
                prefix = "[%s]" % self.name
                with override_log_formatting(prefix + "%(message)s"):
                    for line in logs:
                        print_func(line[:-1].decode("utf-8"))
            else:
                for line in logs:
                    print_func(line[:-1].decode("utf-8"))
        except Exception as ex:
            self.log.error("Failed to stream logs:  %s", self.name, ex)

    def check_deploy_errors(self, pod_v1_resp):
        pod_status = pod_v1_resp.status
        if self.kube_config.check_unschedulable_condition and pod_status.conditions:
            for condition in pod_status.conditions:
                if condition.reason != "Unschedulable":
                    continue
                logger.info("pod is pending because %s" % condition.message)
                if (
                    "Insufficient cpu" in condition.message
                    or "Insufficient memory" in condition.message
                ):
                    if self.kube_config.check_cluster_resource_capacity:
                        kube_resources_checker = DbndKubeResourcesChecker(
                            kube_client=self.kube_client, kube_config=self.kube_config
                        )
                        kube_resources_checker.check_if_resource_request_above_max_capacity(
                            condition.message
                        )

                    self.log.warning("pod is pending because %s" % condition.message)
                else:
                    raise friendly_error.executor_k8s.kubernetes_pod_unschedulable(
                        condition.message
                    )

        if pod_status.container_statuses:
            container_waiting_state = pod_status.container_statuses[0].state.waiting
            if (
                self.kube_config.check_image_pull_errors
                and pod_status.phase == "Pending"
                and container_waiting_state
            ):
                if container_waiting_state.reason == "ErrImagePull":
                    raise friendly_error.executor_k8s.kubernetes_image_not_found(
                        pod_status.container_statuses[0].image,
                        container_waiting_state.message,
                        long_msg=container_waiting_state.reason,
                    )

                if container_waiting_state.reason == "CreateContainerConfigError":
                    raise friendly_error.executor_k8s.kubernetes_pod_config_error(
                        container_waiting_state.message
                    )

    def check_running_errors(self, pod_v1_resp):
        """
        Raise an error if pod in running state with Failed conditions
        """
        pod_status = pod_v1_resp.status
        if not self.kube_config.check_running_pod_errors:
            return
        if pod_status.conditions:
            for condition in pod_status.conditions:
                if condition.type != "Ready":
                    continue
                # We are looking for
                #  {
                #   u"status": u"False",
                #   u"lastProbeTime": None,
                #   u"type": u"Ready",
                #   u"lastTransitionTime": u"2021-01-22T04:54:13Z",
                #  },
                if not condition.status or condition.status == "False":
                    raise friendly_error.executor_k8s.kubernetes_running_pod_fails_on_condition(
                        condition, pod_name=pod_v1_resp.metadata.name
                    )
                return True
        return False

    def wait(self):
        """
        Waits for pod completion
        :return:
        """
        self._wait_for_pod_started()
        self.log.info("Pod is running, reading logs..")
        self.stream_pod_logs(follow=True)
        self.log.info("Successfully read pod logs")

        pod_phase = self.get_pod_phase()
        wait_start = utcnow()
        while pod_phase not in {PodPhase.SUCCEEDED, PodPhase.FAILED}:
            logger.debug(
                "Pod '%s' is not completed with state %s, waiting..",
                self.name,
                pod_phase,
            )
            if (
                utcnow() - wait_start
            ) > self.kube_config.submit_termination_grace_period:
                raise DatabandRuntimeError(
                    "Pod is not in a final state after {grace_period}: {state}".format(
                        grace_period=self.kube_config.submit_termination_grace_period,
                        state=pod_phase,
                    )
                )
            time.sleep(5)
            pod_phase = self.get_pod_phase()

        if pod_phase != PodPhase.SUCCEEDED:
            raise DatabandRuntimeError(
                "Pod returned a failure: {pod_phase}".format(pod_phase=pod_phase)
            )
        return self

    def run_pod(
        self, task_run: "TaskRun", pod: "k8s.V1Pod", detach_run: bool = False
    ) -> "DbndPodCtrl":
        kc = self.kube_config
        detach_run = detach_run or kc.detach_run
        if not self.is_possible_to_detach_run():
            detach_run = False

        req = kc.build_kube_pod_req(pod)
        self._attach_live_logs_container(req)

        readable_req_str = readable_pod_request(req)

        if kc.debug:
            logger.info("Pod Creation Request: \n%s", readable_req_str)
            pod_file = task_run.task_run_attempt_file("pod.yaml")
            pod_file.write(readable_req_str)
            logger.debug("Pod Request has been saved to %s", pod_file)

        external_link_dict = self.build_external_links(pod)
        if external_link_dict:
            task_run.set_external_resource_urls(external_link_dict)

        task_run.set_task_run_state(TaskRunState.QUEUED)

        try:
            resp = self.kube_client.create_namespaced_pod(
                body=req, namespace=pod.metadata.namespace
            )
            logger.info(
                "%s has been submitted at pod '%s' at namespace '%s'"
                % (task_run, pod.metadata.name, pod.metadata.namespace)
            )
            self.log.debug("Pod Creation Response: %s", resp)
        except ApiException as ex:
            task_run_error = TaskRunError.build_from_ex(ex, task_run)
            task_run.set_task_run_state(TaskRunState.FAILED, error=task_run_error)
            logger.error(
                "Exception when attempting to create Namespaced Pod using: %s",
                readable_req_str,
            )
            raise

        if detach_run:
            return self

        self.wait()
        return self

    def _attach_live_logs_container(self, req: typing.Dict[str, typing.Any]):
        from dbnd_docker.kubernetes.vendorized_airflow.request_factory import (
            DbndPodRequestFactory,
        )

        DbndPodRequestFactory(self.kube_config).attach_logs_container(req)

    def build_external_links(self, pod: "k8s.V1Pod"):
        kc = self.kube_config
        dashboard_url = kc.get_dashboard_link(pod.metadata.namespace, pod.metadata.name)
        pod_log = kc.get_pod_log_link(pod.metadata.namespace, pod.metadata.name)
        external_link_dict = dict()
        if dashboard_url:
            external_link_dict["k8s_dashboard"] = dashboard_url
        if pod_log:
            external_link_dict["pod_log"] = pod_log
        return external_link_dict

    def is_possible_to_detach_run(self):
        kc = self.kube_config
        can_detach_run = True
        if kc.show_pod_log:
            logger.info(
                "%s is True, %s will send every docker in blocking mode",
                "show_pod_logs",
                kc.task_name,
            )
            can_detach_run = False
        if kc.debug:
            logger.info(
                "%s is True, %s will send every docker in blocking mode",
                "debug",
                kc.task_name,
            )
            can_detach_run = False
        return can_detach_run

    def get_pod_logs(self, tail_lines=100):
        try:
            logs = []
            log_printer = lambda x: logs.append(x)
            self.stream_pod_logs(
                print_func=log_printer, tail_lines=tail_lines, follow=False
            )
            return logs
        except ApiException as ex:
            if ex.status == 404:
                self.log.info("failed to get log for pod: pod not found")
            else:
                self.log.exception("failed to get log: %s", ex)
        except Exception as ex:
            self.log.error("failed to get log for %s: %s", ex)


def _get_status_log_safe(pod_data):
    try:
        pp = pprint.PrettyPrinter(indent=4)
        logs = pp.pformat(pod_data.status)
        return "Pod Status: %s" % logs
    except Exception as ex:
        return "Pod Status: failed to get %s: %s" % (pod_data.metadata.name, ex)


def _try_get_pod_exit_code(pod_data) -> (str, str):
    found_exit_code = False
    pod_exit_code = None
    termination_reason = None
    if pod_data.status.container_statuses:
        for container_status in pod_data.status.container_statuses:
            # Searching for the container that was terminated
            if container_status.state.terminated:
                pod_exit_code = container_status.state.terminated.exit_code
                termination_reason = container_status.state.terminated.reason
                found_exit_code = True
                if container_status.name == "base":
                    break

        if found_exit_code:
            return pod_exit_code, termination_reason
    return None, None


def get_task_run_from_pod_data(pod_data):
    labels = pod_data.metadata.labels
    if "task_id" not in labels:
        return None
    task_id = labels["task_id"]

    dr = try_get_databand_run()
    if not dr:
        return None

    return dr.get_task_run_by_af_id(task_id)


class PodPhase(object):
    """Status of the PODs"""

    PENDING = "Pending"
    RUNNING = "Running"
    FAILED = "Failed"
    SUCCEEDED = "Succeeded"



import logging
import math

from dbnd._core.errors import DatabandConfigError, friendly_error


logger = logging.getLogger(__name__)

SI_MEMORY_SUFFIXES = ["K", "M", "G", "T", "P", "E"]
ISO_MEMORY_SUFFIXES = ["%si" % s for s in SI_MEMORY_SUFFIXES]
MEMORY_SUFFIXES_WITH_BASE_AND_POWER = [
    (ISO_MEMORY_SUFFIXES, 2, 10),
    (SI_MEMORY_SUFFIXES, 10, 3),
]

LIST_NODES_TIMEOUT_S = 3


class DbndKubeResourcesChecker(object):
    def __init__(self, kube_client, kube_config):
        super(DbndKubeResourcesChecker, self).__init__()
        self.kube_client = kube_client
        self.kube_config = kube_config

    def check_if_resource_request_above_max_capacity(self, condition_message):

        try:
            cpu_request = self.get_cpu_request()
            memory_request = self.get_memory_request()

            if cpu_request or memory_request:
                nodes = self.get_cluster_nodes()
                if not nodes:
                    return

                message_components = []
                if cpu_request:
                    nodes_with_cpu = [
                        (node, float(node.status.capacity["cpu"])) for node in nodes
                    ]
                    max_cpu_node = max(nodes_with_cpu, key=lambda nwc: nwc[1])
                    if cpu_request > max_cpu_node[1]:
                        message_components.append(
                            self.get_resource_above_max_capacity_message(
                                param="request_cpu",
                                param_value=self.kube_config.request_cpu,
                                name="CPUs",
                                max_node=max_cpu_node[0].metadata.name,
                                max_capacity=max_cpu_node[0].status.capacity["cpu"],
                            )
                        )

                if memory_request:
                    nodes_with_memory = [
                        (node, parse_kub_memory_string(node.status.capacity["memory"]))
                        for node in nodes
                    ]
                    max_memory_node = max(nodes_with_memory, key=lambda nwm: nwm[1])
                    if memory_request > max_memory_node[1]:
                        message_components.append(
                            self.get_resource_above_max_capacity_message(
                                param="request_memory",
                                param_value=self.kube_config.request_memory,
                                name="memory",
                                max_node=max_memory_node[0].metadata.name,
                                max_capacity=max_memory_node[0].status.capacity[
                                    "memory"
                                ],
                            )
                        )

                if message_components:
                    raise friendly_error.executor_k8s.kubernetes_pod_unschedulable(
                        condition_message, extra_help="\n".join(message_components)
                    )

        except DatabandConfigError as ex:
            raise ex
        except Exception as ex:
            logger.warning(
                "failed to check if the requested pod resources are above maximum cluster capacity: %s"
                % ex.args
            )

    def get_resource_above_max_capacity_message(self, **kwargs):
        return (
            "Configured {param} ({param_value}) is larger than the node with most {name} in the cluster ({max_node} with {max_capacity}). "
            "Either lower the {param} or add a node with enought {name} to the cluster."
        ).format(**kwargs)

    def get_cpu_request(self):
        """
        https://kubernetes.io/docs/concepts/configuration/manage-compute-resources-container/#meaning-of-cpu
        :return: float value of requested cpu fractions or None
        """
        try:
            if not self.kube_config.request_cpu:
                return None

            if self.kube_config.request_cpu[-1].lower() == "m":
                return int(self.kube_config.request_cpu[:-1]) / 1000.0
            else:
                return float(self.kube_config.request_cpu)
        except ValueError as e:
            raise DatabandConfigError(
                "failed to parse request_cpu %s: %s" % (self.kube_config.request_cpu, e)
            )

    def get_memory_request(self):
        try:
            return parse_kub_memory_string(self.kube_config.request_memory)
        except DatabandConfigError as e:
            raise DatabandConfigError("failed to parse request_memory: %s" % e.args)

    def get_cluster_nodes(self):
        try:
            return self.kube_client.list_node(
                timeout_seconds=LIST_NODES_TIMEOUT_S
            ).items
        except Exception as ex:
            logger.warning(
                "tried to list Kubernetes cluster nodes to generate a better failure message but failed because: %s"
                % ex
            )

        return None


def parse_kub_memory_string(memory_string):
    """
    https://kubernetes.io/docs/concepts/configuration/manage-compute-resources-container/#meaning-of-memory
    :return: float value of requested bytes or None
    """
    if not memory_string:
        return None

    try:
        for suffixes, base, power in MEMORY_SUFFIXES_WITH_BASE_AND_POWER:
            for i, s in enumerate(suffixes, start=1):
                if memory_string.endswith(s) or memory_string.endswith(s.lower()):
                    return float(memory_string[: -len(s)]) * math.pow(base, power * i)
    except ValueError as e:
        raise DatabandConfigError("memory parse failed for %s: %s" % (memory_string, e))

    raise DatabandConfigError(
        "memory parse failed for %s: suffix not recognized" % memory_string
    )



import subprocess

from dbnd import parameter
from dbnd._core.parameter.validators import NonEmptyString
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.settings import EngineConfig
from targets.values.version_value import VersionStr


class ContainerEngineConfig(EngineConfig):
    require_submit = True
    dbnd_executable = ["dbnd"]  # we should have 'dbnd' command installed in container
    container_repository = parameter(validator=NonEmptyString()).help(
        "Docker container registry"
    )[str]
    container_tag = parameter.none().help("Docker container tag")[VersionStr]
    container_tag_gpu = parameter.none().help("Docker container tag for GPU tasks")[
        VersionStr
    ]

    docker_build_tag_base = parameter.help("Auto build docker container tag").value(
        "dbnd_build"
    )
    docker_build_tag = parameter.help(
        "Docker build tag for the docker image dbnd will build"
    ).default(None)[str]
    docker_build = parameter(default=True).help(
        "Automatically build docker image. "
        "If container_repository is unset it will be taken (along with the tag) from the docker build settings"
    )[bool]
    docker_build_push = parameter(default=True).help(
        "If docker_build is enabled, controls whether the image is automatically pushed or not"
    )

    def get_docker_ctrl(self, task_run):
        pass

    @property
    def full_image(self):
        return "{}:{}".format(self.container_repository, self.container_tag)

    def prepare_for_run(self, run):
        # type: (DatabandRun) -> None
        super(ContainerEngineConfig, self).prepare_for_run(run)

        from dbnd_docker.submit_ctrl import prepare_docker_for_executor

        # when we run at submitter - we need to update driver_engine - this one will be used to send job
        # when we run at driver - we update task config, it will be used by task
        # inside pod submission the fallback is always on task_engine

        prepare_docker_for_executor(run, self)

    def submit_to_engine_task(self, env, task_name, args, interactive=True):
        from dbnd_docker.docker.docker_task import DockerRunTask

        submit_task = DockerRunTask(
            task_name=task_name,
            command=subprocess.list2cmdline(args),
            image=self.full_image,
            docker_engine=self,
            task_is_system=True,
        )
        return submit_task

    def _should_wrap_with_submit_task(self, task_run):
        """
        We don't want to resubmit if it's dockerized run and we running with the same engine
        """
        from dbnd_docker.docker.docker_task import DockerRunTask

        if isinstance(task_run.task, DockerRunTask):
            if task_run.task.docker_engine.task_name == self.task_name:
                return False
        return super(ContainerEngineConfig, self)._should_wrap_with_submit_task(
            task_run
        )



from os import path

import setuptools

from setuptools.config import read_configuration


BASE_PATH = path.dirname(__file__)
CFG_PATH = path.join(BASE_PATH, "setup.cfg")

config = read_configuration(CFG_PATH)
version = config["metadata"]["version"]

setuptools.setup(
    name="dbnd-docker",
    package_dir={"": "src"},
    install_requires=[
        "dbnd==" + version,
        "dbnd-airflow==" + version,
        "docker>=3.0",
        # k8s
        "kubernetes>=9.0.0",
        "cryptography>=2.0.0",
    ],
    entry_points={"dbnd": ["dbnd-docker = dbnd_docker._plugin"]},
)



import logging
import os

from typing import Any, Dict


AIRFLOW_LOG_MOUNT_NAME = "airflow-logs"

logger = logging.getLogger(__name__)


class DbndPodRequestFactory(object):
    def __init__(self, kubernetes_engine_config):
        self.kubernetes_engine_config = kubernetes_engine_config

    def attach_logs_container(self, mutable_pod_creation_request: Dict[str, Any]):
        if self.kubernetes_engine_config.airflow_log_enabled:
            self.add_logs_container(mutable_pod_creation_request)

    def add_logs_container(self, req):
        if self.kubernetes_engine_config.trap_exit_file_flag is None:
            logger.warning(
                "Can't use live log feature without trap exit file."
                "Please use the configuration `kubernetes.trap_exit_file_flag`"
            )

        trap_file = self.kubernetes_engine_config.trap_exit_file_flag
        trap_dir = os.path.dirname(trap_file)

        # add a new mount volume for the logs
        req["spec"].setdefault("volumes", [])
        req["spec"]["volumes"].extend(
            [
                {"name": AIRFLOW_LOG_MOUNT_NAME, "emptyDir": {}},
                {"name": "trap", "emptyDir": {}},
            ]
        )

        # add to the log mount to the original container
        base_container = req["spec"]["containers"][0]
        base_container.setdefault("volumeMounts", [])
        base_container["volumeMounts"].extend(
            [
                {
                    "name": AIRFLOW_LOG_MOUNT_NAME,
                    "mountPath": self.kubernetes_engine_config.container_airflow_log_path,
                },
                {"name": "trap", "mountPath": trap_dir},
            ]
        )

        # configure airflow to use the ip as the hostname of the job
        # airflow-webserver uses the hostname to query the log_sidecar but in kubernetes
        # In kubernetes normally, only Services get DNS names, not Pods. so we use the ip.
        # see more here: https://stackoverflow.com/a/59262628
        if self.kubernetes_engine_config.host_as_ip_for_live_logs:
            base_container.setdefault("env", [])
            base_container["env"].append(
                {
                    "name": "AIRFLOW__CORE__HOSTNAME_CALLABLE",
                    "value": "airflow.utils.net:get_host_ip_address",
                }
            )

        # build the log sidecar and add it
        log_folder = self.kubernetes_engine_config.airflow_log_folder

        airflow_image = (
            self.kubernetes_engine_config.airflow_log_image or base_container["image"]
        )
        log_port = self.kubernetes_engine_config.airflow_log_port
        side_car = {
            "name": "log-sidecar",  # keep the name lowercase
            "image": airflow_image,
            "command": ["/bin/bash", "-c"],
            "args": [
                """airflow serve_logs &
                CHILD_PID=$!
                (while true ; do if [[ -f "{0}" ]]; then kill $CHILD_PID; fi; sleep 1; done) &
                wait $CHILD_PID
                if [[ -f "{0}" ]]; then exit 0; fi""".format(
                    trap_file
                )
            ],
            "volumeMounts": [
                {"name": AIRFLOW_LOG_MOUNT_NAME, "mountPath": log_folder},
                {"name": "trap", "mountPath": trap_dir},
            ],
            "ports": [{"containerPort": int(log_port)}],
        }

        # we want our side container to have basic env
        side_car["env"] = (
            list(base_container["env"]) if base_container.get("env") else []
        )
        existing_keys = {env_def.get("name") for env_def in side_car["env"]}

        # those env variable are used by `airflow serve_logs` to access the log and with the relevant port
        if "AIRFLOW__CORE__BASE_LOG_FOLDER" not in existing_keys:
            side_car["env"].append(
                {"name": "AIRFLOW__CORE__BASE_LOG_FOLDER", "value": log_folder}
            )
        if "AIRFLOW__CELERY__WORKER_LOG_SERVER_PORT" not in existing_keys:
            side_car["env"].append(
                {"name": "AIRFLOW__CELERY__WORKER_LOG_SERVER_PORT", "value": log_port}
            )

        req["spec"]["containers"].append(side_car)



import copy

from typing import Any, Dict, Optional, Union

import kubernetes.client.models as k8s


class DbndExtendedResources:
    def __init__(
        self,
        requests: Optional[Dict[str, Any]] = None,
        limits: Optional[Dict[str, Any]] = None,
        request_cpu: Optional[Union[str, float]] = None,
        request_memory: Optional[str] = None,
        limit_cpu: Optional[Union[str, float]] = None,
        limit_memory: Optional[str] = None,
    ):
        self.requests = requests or {}
        self.limits = limits or {}
        self.request_cpu = request_cpu
        self.request_memory = request_memory
        self.limit_cpu = limit_cpu
        self.limit_memory = limit_memory

    def to_k8s_client_obj(self) -> k8s.V1ResourceRequirements:
        limits_raw = {"cpu": self.limit_cpu, "memory": self.limit_memory}
        requests_raw = {"cpu": self.request_cpu, "memory": self.request_memory}

        self.limits.update(self.cleanup_empty_values(limits_raw))
        self.requests.update(self.cleanup_empty_values(requests_raw))
        resource_req = k8s.V1ResourceRequirements(
            limits=self.limits, requests=self.requests
        )
        return resource_req

    def cleanup_empty_values(
        self, resource_spec: Dict[str, Optional[Any]]
    ) -> Dict[str, Any]:
        return {k: v for k, v in resource_spec.items() if v}

    def attach_to_pod(self, pod: k8s.V1Pod) -> k8s.V1Pod:
        cp_pod = copy.deepcopy(pod)
        resources = self.to_k8s_client_obj()
        cp_pod.spec.containers[0].resources = resources
        return cp_pod



import logging

from dbnd import parameter
from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.task.task import Task
from dbnd._core.utils.better_subprocess import run_cmd
from dbnd._core.utils.project.project_fs import project_path


logger = logging.getLogger(__name__)


class DockerBuild(Task):
    use_kaniko = parameter(default=False)[bool]

    docker_file = parameter(default=project_path("Dockerfile"))[str]
    image_name = parameter()[str]
    tag = parameter(default="latest")[str]
    label = parameter(default="")[str]
    push = parameter(default=True)[bool]
    target = parameter(default=None)[str]

    working_dir = parameter(default=None)[str]

    kaniko_command = parameter(default=None)[str]
    context = parameter(default=None)[str]
    destinations = parameter(default=None)[list]
    build_args = parameter(default=None)[list]

    full_image_name = None
    computed_tag = None
    image_name_with_tag = None

    def run(self):
        if self.use_kaniko:
            return self.run_using_kaniko()
        else:
            return self.run_using_docker_build()

    def run_using_docker_build(self):
        if self.tag:
            self.image_name_with_tag = "{}:{}".format(self.image_name, self.tag)
        else:
            self.image_name_with_tag = self.full_image_name

        try:
            cmd = "docker build -t {} -f {} .".format(
                self.image_name_with_tag, self.docker_file
            )
            if self.label:
                cmd = cmd + " --label " + self.label
            if self.target:
                cmd = cmd + " --target " + self.target

            if self.build_args:
                build_args_list = [
                    " --build-arg {}".format(arg) for arg in self.build_args
                ]
                cmd = cmd + "".join(build_args_list)

            logger.info("Running docker build command: `%s`\n\n", cmd)
            cwd = self.working_dir or project_path()
            run_cmd(cmd, shell=True, cwd=cwd)

        except Exception as e:
            logger.error(
                "^^^^^^^^^^^^^^^ SEE DOCKER ERROR MESSAGE ABOVE THIS LINE ^^^^^^^^^^^^^^^\n\n"
            )
            raise DatabandRuntimeError(
                "failed building docker image {}".format(self.image_name_with_tag),
                nested_exceptions=[e],
            )

        if self.push:
            try:
                cmd = "docker push {}".format(self.image_name_with_tag)
                logger.info("Running docker push command: '%s'", cmd)
                run_cmd(cmd, shell=True)

            except Exception as e:
                raise DatabandRuntimeError(
                    "failed to push docker image {}".format(self.image_name_with_tag),
                    nested_exceptions=[e],
                )
        else:
            logger.info("skipping docker push")

        return self.image_name_with_tag

    def run_using_kaniko(self):
        if self.tag:
            self.image_name_with_tag = "{}:{}".format(self.image_name, self.tag)
        else:
            self.image_name_with_tag = self.full_image_name

        command = "{} -c {} -f {}".format(
            self.kaniko_command, self.context, self.docker_file
        )

        if not self.destinations:
            command = command + " --no-push"
        else:
            destination_list = [
                " -d {}".format(destination) for destination in self.destinations
            ]
            command = command + "".join(destination_list)

        if self.build_args:
            build_args_list = [" --build-arg {}".format(arg) for arg in self.build_args]
            command = command + "".join(build_args_list)

        if self.label:
            command = command + " --label " + self.label

        if self.target:
            command = command + " --target " + self.target

        try:
            logger.info("Running build using Kaniko: %s", command)
            run_cmd(command, shell=True, cwd=project_path())
        except Exception as e:
            raise DatabandRuntimeError(
                "failed building docker image {}".format("?"), nested_exceptions=[e]
            )



import logging
import typing

from typing import Optional, Type

from dbnd import dbnd_config, log_metric, override
from dbnd._core.constants import CURRENT_TIME_STR
from dbnd._core.errors.friendly_error.executor_k8s import no_tag_on_no_build
from dbnd._core.plugin.dbnd_plugins import pm
from dbnd_docker.container_engine_config import ContainerEngineConfig
from dbnd_docker.docker.docker_build import DockerBuild
from dbnd_docker.kubernetes.kubernetes_engine_config import KubernetesEngineConfig


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun

logger = logging.getLogger(__name__)


def prepare_docker_for_executor(run, docker_engine):
    # type: (DatabandRun, ContainerEngineConfig) -> Optional[DockerBuild]
    if docker_engine.container_tag:
        logger.info(
            "Omitting docker build due to existing container_tag=%s",
            docker_engine.container_tag,
        )
        log_metric("container_tag", docker_engine.container_tag)
        return None

    config_cls = docker_engine.__class__  # type: Type[ContainerEngineConfig]

    def _set_config(parameter, value):
        # set value in already existing object
        setattr(docker_engine, parameter.name, value)
        dbnd_config.set_parameter(
            parameter, override(value), source="prepare_docker_for_executor"
        )

    if docker_engine.docker_build:
        if docker_engine.docker_build_tag:
            auto_tag = docker_engine.docker_build_tag
        else:
            auto_tag = docker_engine.docker_build_tag_base + "_" + CURRENT_TIME_STR
        _set_config(config_cls.container_tag, auto_tag)

        if (
            isinstance(docker_engine, KubernetesEngineConfig)
            and not docker_engine.docker_build_push
        ):
            _set_config(config_cls.image_pull_policy, "Never")

        log_metric("docker build tag", auto_tag)
        log_metric("container_tag", auto_tag)
        docker_build = DockerBuild(
            task_name="dbnd_image_build",
            image_name=docker_engine.container_repository,
            tag=docker_engine.container_tag,
            push=docker_engine.docker_build_push,
            task_version="now",
            task_is_system=True,
        )

        run.run_executor.run_task_at_execution_time(docker_build)
        pm.hook.dbnd_build_project_docker(
            docker_engine=docker_engine, docker_build_task=docker_build
        )

        return docker_build
    else:
        logger.info("Omitting docker build due to docker_build=False")
        if not docker_engine.container_tag:
            raise no_tag_on_no_build()
    return None
